unit uFeatureParser;

interface

uses
  Classes, DelphiSpec.Parser, DelphiSpec.Scenario;

type
  TMyDelphiSpecParser = class(TDelphiSpecParser)
  public
    procedure Execute(const FileName: string; Features: TFeatureList); override;
  end;

  TFeatureParser = class
  private
    FGeneratedOutput: TStringList;

    procedure RenameExistingOutputfile(const FileName: string);
    procedure GeneratePasHeader(const UnitName: string);

    procedure GenerateInterface(const Features: TFeatureList);
    procedure GenerateImplementation(const Features: TFeatureList);
    procedure GenerateFinalSections(const Features: TFeatureList);
    procedure ClearStatus;
    procedure GenerateInterfaceSteps(Scenario: TScenario);
    procedure GenerateImplementationSteps(Feature: TFeature; Scenario: TScenario);
    function GetProcedureHeader(StepString: string): string;
  public
    constructor Create;
    destructor Destroy; override;

    function Parse(const FilePath: string): Boolean;
  end;

implementation

uses
  System.IOUtils, System.SysUtils, System.RegularExpressions;

const
  FEATURENAME = 'feature';
  STEPNAME: array of string = ['scenario outline', 'scenario', 'given', 'when', 'then', 'and', 'but'];

{ TFeatureParser }

constructor TFeatureParser.Create;
begin
  FGeneratedOutput := TStringList.Create;
end;

destructor TFeatureParser.Destroy;
begin
  FGeneratedOutput.Free;

  inherited;
end;

procedure TFeatureParser.GenerateFinalSections(const Features: TFeatureList);
var
  Feature: TFeature;
begin
  FGeneratedOutput.Add('initialization');
  for Feature in Features do
    FGeneratedOutput.Add(Format('  RegisterStepDefinitionsClass(T%s);', [Feature.Name.Replace(' ', '_')]));
  FGeneratedOutput.Add('');
  FGeneratedOutput.Add('end.');
end;

procedure TFeatureParser.GenerateImplementation(const Features: TFeatureList);
var
  Feature: TFeature;
  Scenario: TScenario;
  ScenarioOutline: TScenarioOutline;
begin
  FGeneratedOutput.Add('implementation');
  FGeneratedOutput.Add('');
  FGeneratedOutput.Add('uses');
  FGeneratedOutput.Add('  DelphiSpec.Core;');
  FGeneratedOutput.Add('');

 for Feature in Features do
  begin
    for Scenario in Feature.Scenarios do
      GenerateImplementationSteps(Feature, Scenario);

    for ScenarioOutline in Feature.ScenarioOutlines do
      GenerateImplementationSteps(Feature, ScenarioOutline);

  end;  
end;

procedure TFeatureParser.GenerateInterface(const Features: TFeatureList);
var
  Feature: TFeature;
  Scenario: TScenario;
  ScenarioOutline: TScenarioOutline;
begin
  FGeneratedOutput.Add('interface');
  FGeneratedOutput.Add('');
  FGeneratedOutput.Add('uses');
  FGeneratedOutput.Add('  DelphiSpec.Attributes,');
  FGeneratedOutput.Add('  DelphiSpec.StepDefinitions,');
  FGeneratedOutput.Add('  DUnitX.TestFramework;');
  FGeneratedOutput.Add('');
  FGeneratedOutput.Add('type');

  for Feature in Features do
  begin
    FGeneratedOutput.Add(Format('  [Feature(''%s'')]', [Feature.Name]));
    FGeneratedOutput.Add(Format('  T%s = class(TStepDefinitions)', [Feature.Name.Replace(' ', '_')]));
    FGeneratedOutput.Add(       '  public');

    for Scenario in Feature.Scenarios do
      GenerateInterfaceSteps(Scenario);
    for ScenarioOutline in Feature.ScenarioOutlines do
      GenerateInterfaceSteps(ScenarioOutline);
  end;
end;

procedure TFeatureParser.GeneratePasHeader(const UnitName: string);
begin
  FGeneratedOutput.Add('unit ' + UnitName.Replace(' ', '') + ';');
  FGeneratedOutput.Add('');
end;

function TFeatureParser.Parse(const FilePath: string): Boolean;
var
  FileName: string;
  Parser: TMyDelphiSpecParser;
  FeatureList: TFeatureList;
begin
  Result := False;
  ClearStatus;

  FileName := IncludeTrailingPathDelimiter(TPath.GetDirectoryName(FilePath)) + 'u' + TPath.GetFileNameWithoutExtension(FilePath) + 'Test.pas';
  if FileExists(FileName) then
    RenameExistingOutputfile(FileName);

  FeatureList := TFeatureList.Create;
  Parser := TMyDelphiSpecParser.Create('EN');
  Parser.Execute(FilePath, FeatureList);
  try
    if FeatureList.Count > 0 then
    begin
      GeneratePasHeader(TPath.GetFileNameWithoutExtension(FileName));

      GenerateInterface(FeatureList);
      GenerateImplementation(FeatureList);

      GenerateFinalSections(FeatureList);

      FGeneratedOutput.SaveToFile(FileName);
      Result := True;
    end;
  finally
    FeatureList.Free;
  end;
end;

function TFeatureParser.GetProcedureHeader(StepString: string): string;
var
  ProcHeader: string;
  I: Integer;
  StepParameters: TMatch;
begin
  StepParameters := TRegEx.Match(StepString, '\<(.*)\>');
  ProcHeader := Format('%s', [TRegEx.Replace(StepString, '\<(.*)\>', '$1').Replace(' ', '_')]);
  if StepParameters.Groups.Count > 1 then
  begin
    ProcHeader := ProcHeader + '(';
    for I := 1 to StepParameters.Groups.Count - 1 do
    begin
      if I > 1 then
        ProcHeader := ProcHeader + '; ';
      ProcHeader := ProcHeader + StepParameters.Groups[I].Value + ': string';
    end;    
    ProcHeader := ProcHeader + ')';
  end;
  ProcHeader := ProcHeader + ';';
  Result := ProcHeader;
end;

procedure TFeatureParser.GenerateImplementationSteps(Feature: TFeature; Scenario: TScenario);
var
  StepList: TScenario.TStep;
  ProcHeader: string;
begin
  for StepList in Scenario._Given do
  begin
    ProcHeader := Format('procedure T%s.%s', [Feature.Name.Replace(' ', '_'), GetProcedureHeader(StepList.Value)]);   
    if not FGeneratedOutput.Text.Contains(ProcHeader) then
    begin
      FGeneratedOutput.Add(ProcHeader);
      FGeneratedOutput.Add('begin');
      FGeneratedOutput.Add('');
      FGeneratedOutput.Add('end;');
      FGeneratedOutput.Add('');
    end;
  end;
  for StepList in Scenario._When do
  begin
    ProcHeader := Format('procedure T%s.%s', [Feature.Name.Replace(' ', '_'), GetProcedureHeader(StepList.Value)]);   
    if not FGeneratedOutput.Text.Contains(ProcHeader) then
    begin
      FGeneratedOutput.Add(ProcHeader);
      FGeneratedOutput.Add('begin');
      FGeneratedOutput.Add('');
      FGeneratedOutput.Add('end;');
      FGeneratedOutput.Add('');
    end;
  end;
  for StepList in Scenario._Then do
  begin
    ProcHeader := Format('procedure T%s.%s', [Feature.Name.Replace(' ', '_'), GetProcedureHeader(StepList.Value)]);   
    if not FGeneratedOutput.Text.Contains(ProcHeader) then
    begin
      FGeneratedOutput.Add(ProcHeader);
      FGeneratedOutput.Add('begin');
      FGeneratedOutput.Add('');
      FGeneratedOutput.Add('end;');
      FGeneratedOutput.Add('');
    end;
  end;
end;

procedure TFeatureParser.GenerateInterfaceSteps(Scenario: TScenario);
var
  StepList: TScenario.TStep;
  StepString: string;
  StepParameters: TMatch;
  Group: TGroup;
begin
  for StepList in Scenario._Given do
  begin
    StepString := TRegEx.Replace(StepList.Value, '\<(.*)\>', '\$$1');
    if not FGeneratedOutput.Text.Contains(Format('[Given_(''%s'')]', [StepString])) then
    begin
      FGeneratedOutput.Add(Format('    [Given_(''%s'')]', [StepString]));
      FGeneratedOutput.Add(       '    procedure ' + GetProcedureHeader(StepList.Value));
      FGeneratedOutput.Add('');
    end;
  end;
  for StepList in Scenario._When do
  begin  
    StepString := TRegEx.Replace(StepList.Value, '\<(.*)\>', '\$$1');
    if not FGeneratedOutput.Text.Contains(Format('[When_(''%s'')]', [StepString])) then
    begin
      FGeneratedOutput.Add(Format('    [When_(''%s'')]', [StepString]));
      FGeneratedOutput.Add(       '    procedure ' + GetProcedureHeader(StepList.Value));
      FGeneratedOutput.Add('');
    end;
  end;
  for StepList in Scenario._Then do
  begin
    StepString := TRegEx.Replace(StepList.Value, '\<(.*)\>', '\$$1');
    if not FGeneratedOutput.Text.Contains(Format('[Then_(''%s'')]', [StepString])) then
    begin
      FGeneratedOutput.Add(Format('    [Then_(''%s'')]', [StepString]));
      FGeneratedOutput.Add(       '    procedure ' + GetProcedureHeader(StepList.Value));
      FGeneratedOutput.Add('');
    end;
  end;
end;

procedure TFeatureParser.ClearStatus;
begin
  FGeneratedOutput.Clear;
end;

procedure TFeatureParser.RenameExistingOutputfile(const FileName: string);
var
  Guid: TGUID;
  NewFileName: string;
begin

  repeat
    CreateGUID(Guid);

    NewFileName := IncludeTrailingPathDelimiter(TPath.GetDirectoryName(FileName)) +
                          tpath.GetFileNameWithoutExtension(FileName) + '_' +  GUIDtoString(GUID);
    if TPath.GetExtension(FileName) <> '' then
      NewFileName := NewFileName + '.' + TPath.GetExtension(FileName);
  until not FileExists(NewFileName);


  RenameFile(FileName, NewFileName);
end;

{ TMyDelphiSpecParser }

procedure TMyDelphiSpecParser.Execute(const FileName: string;
  Features: TFeatureList);
var
  Command, FeatureName: string;
  Feature: TFeature;
begin
  // No inherited!

  FReader.LoadFromFile(FileName);

  while not FReader.Eof do
  begin
    PassEmptyLines;
    CheckEof;

    Command := Trim(FReader.ReadLine);
    if not TDelphiSpecLanguages.CheckStepKind(skFeature, Command, FLangCode) then
      RaiseSyntaxError;

    FeatureName := TDelphiSpecLanguages.GetStepText(skFeature, Command, FLangCode);

    Feature := TFeature.Create(FeatureName, nil);
    Features.Add(Feature);
    FeatureNode(Feature);
  end;

end;


end.
